"""
Name: AF_UCS_Algorithm.py
Authors: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Feb. 12, 2010
---------------------------------------------------------------------------------------------------------------------------------------------------------
AF_UCS Version 1.0: An learning classifier system algorithm implemented with our new attribute tracking and feedback strategy for modeling epistasis and heterogeneity 
data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

#!/usr/bin/env python
#Import modules
from AF_UCS_Constants import *
from AF_UCS_ClassifierSet import *
from AF_UCS_PredictionArray import *
from AF_UCS_TimeKeeper import *
from AF_UCS_SNP_Environment import *
import copy
import random

class AF_UCS:
    def __init__(self, e, outputFile, learnIter, timer, remakeFile):
        """ Initializes the AF_UCS algorithm """
        #Global Parameters
        self.env = e
        self.timer = timer
        self.outFile = outputFile
        self.pop = None
        self.maxPopSize = cons.popSize
        self.remakeFile = remakeFile
        self.badMatchFound = 0 #DEBUG
        
        #Parse the iteration specifying string.
        self.iterList = learnIter.split('.')
        for i in range(len(self.iterList)):
            self.iterList[i] = int(self.iterList[i])
        self.iterLast = self.iterList[len(self.iterList)-1]
        
        #Evaluation Parameters
        self.trackFreq = None
        if int(cons.trackFreq) == 0:
            self.trackFreq = self.env.getNumTrainSamples()
        else:
            self.trackFreq = cons.trackFreq
        
        self.learnTrackOut = None
        
        if self.remakeFile == 'False' or self.remakeFile == False:
            try:
                self.learnTrackOut = open(self.outFile+'_LearnTrack.txt','w')     
                self.learnTrackOut.write("Explore_Iteration\tMacroPopSize\tMicroPopSize\tLocalAccuracy\tGenerality\tWeightedGenerality\tTime(min)\n")
            except IOError, (errno, strerror):
                print ("I/O error(%s): %s" % (errno, strerror))
                raise
        else:
            timer.setTimerRestart(self.remakeFile)
            
            try:
                self.learnTrackOut = open(self.outFile+'_LearnTrack.txt','a')     
            except IOError, (errno, strerror):
                print ("I/O error(%s): %s" % (errno, strerror))
                raise
        
        self.evaluationEnv = None
        self.trainAccuracy  = 0
        self.testAccuracy  = 0   
        
        self.attNoiseLim = 1
        self.lastTrain = 0
            
        
    def runAF_UCS(self):
        """ Runs the initialized AF_UCS algorithm. """
        print "Initializing Algorithm"
        explore = 0
        correct  = [0.0 for i in range(self.trackFreq)]
        if self.remakeFile == False or self.remakeFile == 'False':  #Normal "start from the beginning" run.
            # Instantiate Population
            self.pop = ClassifierSet() #initialization occurs via covering (logical multitasking available to Michigan LCS's)
            exploreIter = 0
        else:   #initialize run from an existing popFile
            #Rebuild existing population from text file.
            self.pop = ClassifierSet(self.remakeFile)
            
            #Extract previous iterations from file name
            temp = self.remakeFile.split('_')
            iterRef = len(temp)-1
            completedIterations = int(temp[iterRef])
            print "Restarting from " +str(completedIterations)+ " iterations."
            
            #Set all iteration values (self.iterList, self.iterLast, and exploreIter)
            exploreIter = completedIterations
            for i in range(len(self.iterList)):
                self.iterList[i] += completedIterations
            self.iterLast += completedIterations
            
        #Run Loop
        while exploreIter < self.iterLast:
            #Major Algorithm Learning Methods

            state_class = self.env.getTrainState(percent=float(exploreIter)/self.iterLast) #ALTERATION?: make it so each instance is used for both an explore and exploit cycle.
            self.doEvolve(state_class, exploreIter, correct)
            #--------------------------------------------------------------
            
            #Track Learning
            self.timer.startTimeEvaluation()
            if (exploreIter%self.trackFreq) == (self.trackFreq - 1) and exploreIter > 0:
                if cons.defaultRuleDelete: #Delete Default rules
                    self.pop.deleteDefaults()    

                self.pop.runGenSum(self.env)
                tempAcc = sum(correct)/float(self.trackFreq) #Accuracy over the last "trackFreq" number of iterations.
                trackString = self.pop.getPopTrack(tempAcc, exploreIter+1, self.timer) # Calculates tracking parameters and returns a string of tracking parameters.
                self.learnTrackOut.write(trackString)
            
            #Complete Population Evaluation(s)        !!!! CONSIDER: whether self.env.DataCount is reset here or not, and if you have to go back to where it was???      
            if exploreIter + 1 in self.iterList:
                print "Running Complete Population Evaluation at " + str(exploreIter + 1)+ " iterations."
                self.evaluationEnv = copy.copy(self.env)  # copy environment for evaluation so as not to interupt learning
                if self.evaluationEnv is self.env:
                    print "Environment Object Error"
                    
                if cons.defaultRuleDelete: #Delete Default rules
                    self.pop.deleteDefaults()
                
                trainEval = []
                testEval = []
                if cons.crossVal: #if cross validation is employed
                    trainEval = self.doPopEvaluation(True)
                    testEval = self.doPopEvaluation(False)
                else:
                    trainEval = self.doPopEvaluation(True)
                    testEval = None
                self.timer.endGlobalTimer()
                self.pop.runGenSum(self.env)
                self.pop.writePopTrack(self.outFile, trainEval, testEval, self.env, self.timer, exploreIter)
                self.pop.writePop(self.outFile, self.env, exploreIter)
                #Co_occurence file generation.
                self.pop.attCo_Occurence(self.outFile, self.env, exploreIter)
                
                #Attribute Tracking - Save to File
                if cons.attributeTracking:
                    self.save_tracking(exploreIter)
                    self.save_tracking_Div(exploreIter)
                print "Back to Evolution"
                
            exploreIter += 1 # only increments every other run(during explore phases)
            self.timer.stopTimeEvaluation() 
            
            self.env.resetState(True) #move to next sample in training set.*****KEY

        try: #Close the Learn Tracking File (open and written to, during the entire run of the algorithm)
            self.learnTrackOut.close()
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise   
        
        print "Algorithm Run Complete"
        
        
    def doEvolve(self, state_class, exploreIter, correct):
        """ Run explore cycle. """
        self.timer.startTimeMatchSet()
        matchSet = ClassifierSet(state_class, self.pop, exploreIter, self.timer)
        self.timer.stopTimeMatchSet()
        
        #For learn tracking progress only################################
        predictionArray = PredictionArray(matchSet, cons.numClasses)
        classSelection = predictionArray.bestClassWinner()
        
        if classSelection == None or classSelection == 'tie':
            classSelection = random.choice([0,1]) 
            
        if classSelection != None:
            if int(classSelection) == int(state_class[1]):
                correct[exploreIter%self.trackFreq] = 1
            else:
                correct[exploreIter%self.trackFreq] = 0
        ################################################################

        correctSet = ClassifierSet(matchSet, int(state_class[1]))
        correctSet.updateCorrectTrack()
        
        #ATTRIBUTE TRACKING - Ambrose###############
        if cons.attributeTracking:
            self.env.updateTrainAttTrack(correctSet, matchSet)
        ############################################
        
        matchSet.updateSet()
        if cons.sub:
            self.timer.startTimeSubsumption()
            correctSet.doCorrectSetSubsumption()
            self.timer.stopTimeSubsumption()
            
        self.timer.startTimeGA()
        #For timed application of AttTrack
        percent = exploreIter/float(self.iterLast)
        
        correctSet.runGA(exploreIter, state_class[0], self.env, self.timer, percent)
        self.timer.stopTimeGA()
        
        
    def doPopEvaluation(self, isTrain):
        """ Performs evaluation of pop via the copied environment - either train or test. """
        #Reset Eval counts to zero
        for i in range(len(self.pop.clSet)):
            if isTrain:
                self.pop.clSet[i].clearTrainEval()
            else:
                self.pop.clSet[i].clearTestEval()
            
        if isTrain:
            samples = self.evaluationEnv.getNumTrainSamples()
        else:
            samples = self.evaluationEnv.getNumTestSamples()
            
        self.evaluationEnv.resetDataCount()
        correct  = [0.0 for i in range(samples)] 
        noMatch = [0.0 for i in range(samples)]
        tieVote = 0
        #Balanced Accuracy Calculation
        TP = 0 #accurate disease class
        TN = 0 #accurate control class
        FP = 0
        FN = 0
        
        for each in range(samples):
            if isTrain:
                state_class = self.evaluationEnv.getTrainState(True)
            else:
                state_class = self.evaluationEnv.getTestState()


            matchSet = ClassifierSet("Evaluation", state_class[0], self.pop)#finish
            
            if matchSet.getSize() == 0:
                noMatch[each%samples] = 1
            else:
                noMatch[each%samples] = 0
        
            predictionArray = PredictionArray(matchSet, cons.numClasses)
            classSelection = predictionArray.bestClassWinner()

            if classSelection == None or classSelection == 'tie':
                #print 'tie or failed prediction'
                if classSelection == 'tie':
                    tieVote += 1
                classSelection = random.choice([0,1]) 
            else:
                if int(classSelection) == int(state_class[1]):
                    correct[each%samples] = 1
                    if int(state_class[1]) == 1: #positive is case negative is control
                        TP += 1
                    else:
                        TN += 1
                else:
                    correct[each%samples] = 0
                    if int(state_class[1]) == 1:
                        FP += 1
                    else:
                        FN += 1
                    
            if int(classSelection) == int(state_class[1]):
                matchSet.updateEval(isTrain,True)
            else:
                matchSet.updateEval(isTrain,False)
                
            self.evaluationEnv.resetState(isTrain)   #next sample
             
        unmatched = sum(noMatch)
        #tempAcc = float(sum(correct))/samples   #standard accuracy
        print "TP is "+str(TP)
        print "TN is "+str(TN)
        print "FP is "+str(FP)
        print "FN is "+str(FN)
        
        #balanced Accuracy
        if TP+FN == 0 and TN+FP == 0:
            print "This should not happen EVER!!!!"
            tempAcc = 0
        elif TP+FN == 0:
            print "Should be rare"
            tempAcc = (TN/float(TN+FP))/float(2)
        elif TN+FP == 0:
            print "Should be rare"
            tempAcc = (TP/float(TP+FN))/float(2)
        else:
            tempAcc = ((TP/float(TP+FN))+(TN/float(TN+FP)))/float(2)
            
        coverage = 1 - (float(unmatched+tieVote)/samples)
        newAccuracy = tempAcc*coverage+(1-coverage)*0.5
        if isTrain:
            type = "Training"
        else:
            type = "Testing"
                
        print "Unmatched = " +str(unmatched)
        print "Vote Ties = " +str(tieVote)
        print "Coverage = " + str(coverage)
         
        print str(sum(correct)) + ' out of '+ str(samples)+' correctly classified.  '+type+' Accuracy = '+str(newAccuracy)
        resultList = [newAccuracy, sum(noMatch)/float(samples)]   
        return resultList
    
    
    def getAttWeight(self, sample, env, attDiv):
        trackList = []
        if cons.attributePunish:
            if attDiv:
                tempList = copy.deepcopy(self.env.fTrainAttTrack[sample])
                tempListCount = copy.deepcopy(self.env.fTrainAttCount[sample])
                tempListPun = copy.deepcopy(self.env.fTrainAttPunTrack[sample])
                tempListCountPun = copy.deepcopy(self.env.fTrainAttPunCount[sample])
                corList = []
                punList = []
                for i in range(len(tempList)):
                    if tempListCount[i] == 0:
                        corList.append(0)
                    else:
                        corList.append(tempList[i]/float(tempListCount[i]))
                for i in range(len(tempListPun)):
                    if tempListCountPun[i] == 0:
                        punList.append(0)
                    else:
                        punList.append(tempListPun[i]/float(tempListCountPun[i]))
                for i in range(len(corList)):
                    trackList.append(corList[i]-punList[i])
            else:
                tList = copy.deepcopy(self.env.fTrainAttTrack[sample]) 
                tList2 = copy.deepcopy(self.env.fTrainAttPunTrack[sample])
                for i in range(len(tList)):
                    trackList.append(tList[i]-tList2[i])
        else:
            if attDiv:
                tempList = copy.deepcopy(self.env.fTrainAttTrack[sample])
                tempListCount = copy.deepcopy(self.env.fTrainAttCount[sample])
                for i in range(len(tempList)):
                    if tempListCount[i] == 0:
                        trackList.append(0)
                    else:
                        trackList.append(tempList[i]/float(tempListCount[i]))
            else:
                trackList = copy.deepcopy(self.env.fTrainAttTrack[sample])
        return trackList
        

    def save_tracking(self, exploreIter):
        """ Ambrose: Prints out Attribute Tracking to txt file. """    
        try:  
            f = open(self.outFile + '_'+ str(exploreIter + 1)+'_AttTrack.txt','w') # Outputs tab delimited text file of rule population and respective rule stats
            print "Writing Attribute Tracking as Data File"
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise

        if cons.attributeID:
            i = 0
            f.write('SampleID'+ '\t')
            for att in self.env.headerList:
                if i < len(self.env.headerList)-1:
                    f.write(str(att) + '\t')
                else:
                    f.write(str(att) + '\n')
                i += 1
       
            ############################################

            for i in range(len(self.env.fTrainAttTrack)):
                trackList = self.getAttWeight(i, self.env, False) #no attDiv
                f.write(str(self.env.attributeIDList[i]) + '\t')
                for attr in trackList:
                    f.write(str(attr) + '\t')
                f.write(str(self.env.fTrainData[i][1])+'\n')

        else:
            #Write File Header#######################
            i = 0
            for att in self.env.headerList:
                if i < len(self.env.headerList)-1:
                    f.write(str(att) + '\t')
                else:
                    pass
                i+=1
            i = 0
            for att in self.env.headerList:
                if i < len(self.env.headerList)-1:
                    f.write(str(att) + '\t')
                else:
                    f.write(str(att) + '\n')
                i+=1        
            ############################################
            i = 0
            for counts in self.env.fTrainAttTrack:
                for attr in counts:
                    f.write(str(attr) + '\t')
                for attr in self.env.fTrainData[i][0]:
                    f.write(str(attr) + '\t')
                f.write(str(self.env.fTrainData[i][1]))
                f.write('\n')
                i+=1

        try:
            f.close() 
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise   
          
          
    def save_tracking_Div(self, exploreIter):
        """ Ambrose: Prints out Attribute Tracking to txt file. """    
        try:  
            f = open(self.outFile + '_'+ str(exploreIter + 1)+'_AttDivTrack.txt','w') # Outputs tab delimited text file of rule population and respective rule stats
            print "Writing Attribute Tracking as Data File"
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        
        if cons.attributeID:
            i = 0
            f.write('SampleID'+ '\t')
            for att in self.env.headerList:
                if i < len(self.env.headerList)-1:
                    f.write(str(att) + '\t')
                else:
                    f.write(str(att) + '\n')
                i += 1
       
            ############################################

            for i in range(len(self.env.fTrainAttTrack)):
                trackList = self.getAttWeight(i, self.env, True) #no attDiv
                f.write(str(self.env.attributeIDList[i]) + '\t')
                for attr in trackList:
                    f.write(str(attr) + '\t')
                f.write(str(self.env.fTrainData[i][1])+'\n')

        else:
            #Write File Header#######################
            i = 0
            for att in self.env.headerList:
                if i < len(self.env.headerList)-1:
                    f.write(str(att) + '\t')
                else:
                    pass
                i+=1
            i = 0
            for att in self.env.headerList:
                if i < len(self.env.headerList)-1:
                    f.write(str(att) + '\t')
                else:
                    f.write(str(att) + '\n')
                i+=1        
            ############################################
            i = 0
            for counts in self.env.fTrainAttTrack:
                for attr in counts:
                    f.write(str(attr) + '\t')
                for attr in self.env.fTrainData[i][0]:
                    f.write(str(attr) + '\t')
                f.write(str(self.env.fTrainData[i][1]))
                f.write('\n')
                i+=1

        try:
            f.close() 
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise   
          
            
    def redundantRuleCheck(self):
        """ DEBUG ONLY """
        newList = []
        matchCount = 0
        for i in range(len(self.pop.clSet)-1):
            for j in range(i+1,len(self.pop.clSet)):
                if self.pop.clSet[i].equals(self.pop.clSet[j]):
                    matchCount += 1
                    #print "matches"
                    #print "Found matching classifier: " +str(self.pop.clSet[j].condition)
                    #print str(self.pop.clSet[i].condition)
            #print self.pop.clSet[i].condition
        print "Illegally Matching Classfiers Found = " +str(matchCount)
        if matchCount > 0:
            self.badMatchFound += 1
