"""
Name: AF_UCS_ClassifierSet.py
Authors: Ryan Urbanowicz - written at Dartmouth College, Hanover, NH, USA
Contact: ryan.j.urbanowicz@darmouth.edu
Created: Feb. 12, 2010
---------------------------------------------------------------------------------------------------------------------------------------------------------
AF_UCS Version 1.0: An learning classifier system algorithm implemented with our new attribute tracking and feedback strategy for modeling epistasis and heterogeneity 
data mining and classification tasks.
Copyright (C) 2013 Ryan Urbanowicz 
This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABLILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, 
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
---------------------------------------------------------------------------------------------------------------------------------------------------------
"""

#!/usr/bin/env python
#Import modules
from AF_UCS_Constants import *
from AF_UCS_Classifier import *
from random import *
import sys

class ClassifierSet:
    #Set Constructors-------------------------------------------------------------------------------------------------------   
    def __init__(self, a=None, b=None, c= None, d = None):
        """ Overloaded initialization: Handles creation of different sets required by algorithm. (i.e. Population, match set, action set) """
        self.parentSet = None
        self.clSet = [] #list of classifiers/rules
        self.numerositySum = 0
        
        #Evaluation Parameters
        self.generality = 0.0
        self.weightedGenerality = 0.0
        self.attributeSpecList = []
        self.attributeAccList = []
        
        if a==None:
            self.initPop()
        elif isinstance(a,str) and b==None and c==None:
            self.rebuildPop(a)
        elif isinstance(a, list) and isinstance(b, ClassifierSet):
            self.matchSet(a,b,c,d)
        elif isinstance(a,ClassifierSet):
            self.correctSet(a,b)
        elif isinstance(a,str) and isinstance(b,list) and isinstance(c, ClassifierSet):
            self.evalMatchSet(b,c)
        else:
            print "ClassifierSet: Set arguments not found."
            
        
    def initPop(self):
        """ Instantiates the starting population. """
        self.numerositySum = 0
        self.parentSet     = None
        
        
    def rebuildPop(self, remakeFile):
        """ Remakes a previously evolved population from a text file. """
        self.numerositySum = 0
        self.parentSet     = None
        
        print "Rebuilding a previous population from: " + str(remakeFile)+"_RulePop.txt"
        
        #*******************Initial file handling**********************************************************
        try:       
            datasetList = []
            f = open(remakeFile+"_RulePop.txt", 'r')
            self.headerList = f.readline().rstrip('\n').split('\t')   #strip off first row
            for line in f:
                lineList = line.strip('\n').split('\t')
                datasetList.append(lineList)
            f.close()
            
        except IOError, (errno, strerror):
            print ("Could not Read Remake File!")
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        except ValueError:
            print ("Could not convert data to an integer.")
            raise
        except:
            print ("Unexpected error:", sys.exc_info()[0])
            raise
        #**************************************************************************************************
        
        for each in datasetList:
            oldCl = Classifier(each)
            self.remakeClassifier(oldCl)
            #incrementally update numerositySum
            numerosityRef = cons.environmentObject.numAttributes + 5
            self.numerositySum += int(each[numerosityRef])
            
            
    def matchSet(self, state_class, pop, exploreIter, timer):
        """ Constructs a match set out of the population. After the creation, it is checked if the match set covers all possible actions
        in the environment. If one or more actions are not present, covering occurs, generating the missing action(s). If maximal
        population size is reached when covering, deletion occurs. """ #checked for numerosity error
        self.parentSet     = pop
        self.numerositySum = 0
        state = state_class[0]
        classification = state_class[1]
        #print state_class #debugg - looks like the same data point is coming through each time.!!!!
        
        #Covering check: Twofold (1)checks that a match is present, and (2) that at least one match dictates the correct class.
        correctClassCovered = False
        
        #Matching
        for i in range(len(pop.clSet)):
            cl = pop.clSet[i]
            if cl.match(state):  #checks for a match
                self.addClassifier(cl)
                if cl.getClass() == classification: #!!!!!!!!!!!!!!was getAction()
                    correctClassCovered = True
                    
        while True:
            again=False
            #Covering
            if not correctClassCovered: 
                if cons.cover == "smart":
                    newCl = Classifier(self.numerositySum+1, exploreIter, state, classification)
                    self.addClassifier(newCl)
                    pop.addClassifier(newCl)
                    #print "Covering Activated" # debugging
                else: 
                    print "Requested covering mechanism not available."
                    
            #Deletion
            timer.startTimeDeletion()
            while pop.numerositySum > cons.popSize:
                cdel = pop.deleteFromPopulation()
                #Checks if deleted rule is in match set.
                pos = self.containsClassifier(cdel)
                if cdel != None and pos != -1: #If it is in match set
                    self.numerositySum = self.numerositySum - 1
                    if cdel.getNumerosity()==0:
                        deletedClass = cdel.getClass()
                        self.removeClassifier(pos)
                        if not correctClassCovered and int(deletedClass) == int(classification):
                            again = True
                            
            timer.stopTimeDeletion()
            if not again:
                break
        
        
    def correctSet(self, matchSet, classification):
        """ Constructs a correct set out of the given match set. """  
        self.parentSet = matchSet
        self.numerositySum = 0
        
        for i in range(len(matchSet.clSet)):
            if int(matchSet.clSet[i].getClass()) == int(classification):
                self.addClassifier(matchSet.clSet[i])
                
        
    def evalMatchSet(self, state, pop):
        """ Constructs a match set which does not include covering or deletion, so as not to disrupt the existing population
        and allow for it's evaluation in a static manner as an evolved solution. """
        self.parentSet = pop
        self.numerositySum = 0
               
        for i in range(len(pop.clSet)):
            cl = pop.clSet[i]
            if cl.match(state):
                self.addClassifier(cl)        
    #-----------------------------------------------------------------------------------------------------------------    
    
    
    def deleteDefaults(self):
        """Delete any rules that are completely general """
        keepGoing = True
        counter = len(self.clSet)-1
        while keepGoing and counter > -1:
            cl = self.clSet[counter]
            if cl.isAllWild():
                print "Notice: Default Rule Identified and Removed."
                #print cl.numerosity
                keepGoing = False
                self.numerositySum -= cl.numerosity
                self.removeClassifier(counter)
            counter -= 1
            
            
    def containsClassifier(self,clP):
        """ Returns the position of the classifier in the set if it is present and -1 otherwise. Modified for python lists. """ 
        try:
            return self.clSet.index(clP)
        except ValueError:
            return -1
        
        
    def remakeClassifier(self, classifier):
        """ Adds new classifier object to the population. """
        self.clSet.append(classifier)
        
        
    def addClassifier(self, classifier):
        """ Adds a classifier to the set and increases the numerositySum value accordingly.
        @param classifier The to be added classifier. """

        self.clSet.append(classifier)
        self.addValues(classifier)
        
        
    def addValues(self,cl):
        """ Increases the numerositySum value with the numerosity of the classifier. """
        self.numerositySum = self.numerositySum + cl.getNumerosity()
        
        
    def deleteFromPopulation(self):
        """ Deletes one classifier in the population.  The classifier that will be deleted is chosen by roulette wheel selection
        considering the deletion vote. Returns the macro-classifier which got decreased by one micro-classifier. """
        meanFitness = self.getFitnessSum()/float(self.numerositySum)
        sum = 0.0
        for cl in self.clSet:
            sum = sum + cl.getDelProp(meanFitness)

        choicePoint = sum*random()
        sum=0.0
        i = 0
        for cl in self.clSet:
            sum = sum + cl.getDelProp(meanFitness)
            if sum > choicePoint:
                cl.addNumerosity(-1)
                self.numerositySum = self.numerositySum-1
                if cl.getNumerosity()==0:
                    self.removeClassifier(i)
                return cl
            i = i + 1
        return None
    
    #Remove Classifier----------------------------------------------------------------------
    def removeClassifier(self,param=None):
        """ overloaded method see removeClassifier1(classifier), removeClassifier2(pos)"""
        if isinstance(param,Classifier):
            self.removeClassifier1(param)
        else:
            self.removeClassifier2(param)
        
        
    def removeClassifier1(self, classifier):
        """ Removes the specified (possible macro-) classifier from the population.
        The function returns true when the classifier was found and removed and false
        otherwise. It does not update the numerosity sum of the set, neither
        recursively remove classifiers in the parent set. This must be done manually where required.
        Hard modified for Pythonic way of doing things with lists. """
        try:
            self.clSet.remove(classifier)
            return True
        except ValueError:
            raise

        
    def removeClassifier2(self,pos):
        """ Removes the (possible macro-) classifier at the specified array position from the population.
        The function returns true when the classifier was found and removed and false
        otherwise. It does not update the numerosity of the set, neither
        recursively remove classifiers in the parent set. This must be done manually where required.
        Hard modified for Pythonic way of doing things with lists. """
        try:
            del self.clSet[pos]
        except IndexError:
            raise
        return True
    #---------------------------------------------------------------------------------------
    
    def getFitnessSum(self):
        """ Returns the sum of the fitnesses of all classifiers in the set. """
        sum=0.0
        for cl in self.clSet:
            sum = sum + cl.getFitness()
        return sum
    

    def updateCorrectTrack(self):
        """ Added for UCS to manage correct classifications """
        for cl in self.clSet:
            cl.increaseCorrectTrack()
    

    def updateSet(self):
        """ Updates all parameters in the current set (should be the match set). """
        for cl in self.clSet:
            cl.increaseExperience()            
            cl.updateAccuracy()
            cl.updateCorrectSetSize(self.numerositySum)
            cl.updateFitness()
            
            
    def updateEval(self,isTrain,action):
        """ Testing and Training Match and Correct Tracking for Rules """  #note make sure to implement this post rule compaction as well.
        for cl in self.clSet:
            cl.increaseMatch(isTrain)
            if int(action) == int(cl.classification):
                cl.increaseCorrect(isTrain)
        
        
    def doCorrectSetSubsumption(self):
        """ Executes correct set subsumption.  The correct set subsumption looks for the most general subsumer classifier in the correct set
        and subsumes all classifiers that are more specific than the selected one. This does not appear to remove classifiers from whole pop, 
        only from the correct set - leading to altered GA options."""
        # Recurses back to reference of entire population
        pop = self
        while pop.parentSet!=None:
            pop=pop.parentSet
        
        subsumer = None
        for cl in self.clSet:
            if cl.isSubsumer():
                if subsumer==None or cl.isMoreGeneral(subsumer):
                    subsumer=cl

        #If a subsumer was found, subsume all more specific classifiers in the correct set
        if subsumer!=None:
            i=0
            while i < len(self.clSet):
                if subsumer.isMoreGeneral(self.clSet[i]): #!!!!!!!!!!!!ALSO NEED TO ADD FITNESS CAVIAT
                    num = self.clSet[i].getNumerosity()
                    subsumer.addNumerosity(num)
                    self.clSet[i].addNumerosity((-1)*num)
                    pop.removeClassifier(self.clSet[i]) # AND THIS !!!!!!!!!!!!!!!!!!!!
                    self.removeClassifier(i)
                    i = i-1 
                i = i + 1
    
    
    #GA-call-based----------------------------------------------------------------------
    def runGA(self, exploreIter, state, env, timer, percent):
        """ The Genetic Discovery in AF_UCS takes place here. """
        # Don't do a GA if the theta_GA threshold is not reached - Effect: Initially GA won't be applied, and later if rules were activated recently they will not have GA applied.
        if len(self.clSet) < 1 or (exploreIter - self.getIterStampAverage()) < cons.theta_GA:  
            return
    
        self.setIterStamps(exploreIter) #Show that all included rules have been involved with GA in current iteration.
        
        #Selection Method
        timer.startTimeSelection()

        if cons.sel == "roulette": #Roulette Wheel Selection
            fitSum =self.getFitnessSum()
            cl1P = self.selectClassifierRW(fitSum)
            cl2P = self.selectClassifierRW(fitSum)  
        elif cons.sel == "tournament":
            cl1P = self.selectClassifierT()
            cl2P = self.selectClassifierT()
        else:
            print "Requested GA selection method not available!"

        timer.stopTimeSelection()
        
        # Create offspring copies of the parents which have 0 experience and 1 numerosity.
        cl1  = Classifier(cl1P, exploreIter)
        cl2  = Classifier(cl2P, exploreIter)
        changed = False
        #Crossover
        if cons.cross != "none" and not cl1.equals(cl2) and random() < cons.pX:
            timer.startTimeCrossover()
            #Crossover Method
            if cons.cross == "spc":
                changed = cl1.singlePointCrossover(cl2)
            elif cons.cross == "dpc":
                changed = cl1.twoPointCrossover(cl2)
            elif cons.cross == "uc":
                changed = cl1.uniformCrossover(cl2, env, percent)
            elif cons.cross == "huc":
                changed = cl1.halfUniformCrossover(cl2)
            else:
                print "Requested crossover method not available!"

            #Set Offspring parameters from crossover
            if changed:
                cl1.setAccuracy((cl1.getAccuracy() + cl2.getAccuracy())/2.0)
                cl1.setFitness(cons.fitnessReduction * (cl1.getFitness() + cl2.getFitness())/2.0)
                cl2.setAccuracy(cl1.getAccuracy())
                cl2.setFitness(cl1.getFitness())
            else:   # ensure that the fitness discount is still applied for having run the GA at all
                cl1.setFitness(cons.fitnessReduction * cl1.getFitness())
                cl2.setFitness(cons.fitnessReduction * cl2.getFitness())
            timer.stopTimeCrossover()
            
        nowchanged = False
        howboutnow = False
        if cons.mut: 
            timer.startTimeMutation()
            nowchanged = cl1.applyMutation(state, env, percent)
            howboutnow = cl2.applyMutation(state, env, percent)
            timer.stopTimeMutation()
            
        if changed or nowchanged or howboutnow:
            self.insertDiscoveredClassifiers(cl1, cl2, cl1P, cl2P, timer)    
            
        #PREDICTIVE RULE LOCAL SEARCH
        if cons.p_rule_search != False and random() < cons.p_rule_search: #Activate Predictive Rule Local Search
            print "Running Local Predictive Rule Search at " + str(cons.p_rule_search)+" and " + str(exploreIter)
            if cons.sel == "roulette": #Roulette Wheel Selection
                fitSum =self.getFitnessSum()
                clP = self.selectClassifierRW(fitSum)
            elif cons.sel == "tournament":
                clP = self.selectClassifierT()
            else:
                print "Requested GA selection method not available!"
                
            specifiedCount = 0
            localCount = None

            for j in range(len(clP.condition )):
                if clP.condition[j] != cons.wild:
                    specifiedCount += 1
                        
            print clP.condition # DEBUG
            
            #Look for a minimum of 9 neighbors - complete local 2-way search of genotype combinations
            if specifiedCount == 0:
                return
            if specifiedCount == 1:
                localCount = 3
            else:
                localCount = 9

            localList = []
            localList.append(Classifier(clP,exploreIter)) #put the parent in the list in the first position
            counter = 1
            while len(localList) < localCount:
                localList.append(Classifier(clP,exploreIter))
                localList[counter].makeLocalCondition()
                
                good = True
                i = 0
                while good and i < len(localList)-1:
                    if localList[counter].equals(localList[i]): #If its the same condition
                        localList.pop(counter)
                        good = False
                        counter -= 1
                    i += 1
                counter += 1 
                
            for j in range(len(localList)): #Make classifiers with other class.
                localList.append(Classifier(localList[j],exploreIter)) #make parent out of parent in list
                localList[len(localList)-1].switchClass()
            
            self.insertLocalClassifiers(localList) #Make this method in this class
            
            
    def insertLocalClassifiers(self, localList):
        """  """
        pop = self
        while pop.parentSet!=None: #This works in Python exactly as in Java
            pop = pop.parentSet;

        for i in range(1,len(localList)):
            pop.addClassifierToPopulation(localList[i])
            
        while pop.numerositySum > cons.popSize:
            pop.deleteFromPopulation()
            
            
    def getIterStampAverage(self):
        """ Returns the average of the time stamps in the set. """
        return self.getIterStampSum()/self.numerositySum
    
    
    def getIterStampSum(self):
        """ Returns the sum of the time stamps of all classifiers in the set. """
        sum=0.0
        for cl in self.clSet:
            #print cl.getTimeStamp() #debug
            sum = sum + cl.getTimeStamp() * cl.getNumerosity()
        return sum


    def setIterStamps(self, exploreIter):
        """ Sets the time stamp of all classifiers in the set to the current time. The current time
        is the number of exploration steps executed so far.  """
        for cl in self.clSet:
            cl.setTimeStamp(exploreIter)
        
        
    def selectClassifierRW(self, fitSum):
        """ Selects one classifier using roulette wheel selection according to the fitness of the classifiers. """
        choiceP = random()*fitSum
        i=0
        sum = self.clSet[i].getFitness()
        while choiceP>sum:
            i=i+1
            sum = sum + self.clSet[i].getFitness()
        return self.clSet[i]


    def selectClassifierT(self):
        """  Selects one classifier using tournament selection according to the fitness of the classifiers. """
        actionSetSize = len(self.clSet) 
        tSize = int(actionSetSize*cons.thetaSelect) # sets the number of items in the action set to be included in the tournament selection
        posList = []
        for i in range(tSize): #hold onto a list of random positions, then select the position with the highest fitness
            pos = randrange(actionSetSize)
            if pos in posList: # make sure that pos is a pos that has not yet been selected.
                pos = randrange(actionSetSize)
            else:
                posList.append(pos)
                
        bestF = 0
        bestC = 0
        for j in posList:
            if self.clSet[j].getFitness() > bestF:
                bestF = self.clSet[j].getFitness()
                bestC = j

        return self.clSet[bestC]
        
        
    def insertDiscoveredClassifiers(self, cl1, cl2, cl1P, cl2P, timer):
        """ Inserts both discovered classifiers keeping the maximal size of the population and possibly doing GA subsumption. """
        # Recurses back to reference of entire population
        pop = self
        while pop.parentSet!=None: #This works in Python exactly as in Java
            pop = pop.parentSet;

        if cons.sub:
            timer.startTimeSubsumption()
            self.subsumeClassifier(cl1, cl1P, cl2P)
            self.subsumeClassifier(cl2, cl1P, cl2P)
            timer.stopTimeSubsumption()
        else:
            pop.addClassifierToPopulation(cl1)
            pop.addClassifierToPopulation(cl2)
        
        timer.startTimeDeletion()
        while pop.numerositySum > cons.popSize:
            pop.deleteFromPopulation()
        timer.stopTimeDeletion()
        
        
    def subsumeClassifier(self, cl=None, cl1P=None, cl2P=None):
        """ Tries to subsume a classifier in the parents. If no subsumption is possible it tries to subsume it in the current set. """
        if cl1P!=None and cl1P.subsumes(cl):
            self.increaseNumerositySum(1)
            cl1P.addNumerosity(1)
        elif cl2P!=None and cl2P.subsumes(cl):
            self.increaseNumerositySum(1)
            cl2P.addNumerosity(1)
        else:
            self.subsumeClassifier2(cl); #calls second subsumeUClassifier fkt!

        
    def subsumeClassifier2(self, cl):
        """ Tries to subsume a classifier in the current set. This method is normally called in an action set.
        If no subsumption is possible the classifier is simply added to the population considering
        the possibility that there exists an identical classifier. """
        #Open up a new Vector in order to chose the subsumer candidates randomly
        choices = []
        for cls in self.clSet:
            if cls.subsumes(cl):
                choices.append(cls)

        if len(choices)>0:
            choice = int(random()*len(choices))
            choices[choice].addNumerosity(1)
            self.increaseNumerositySum(1)
            return

        #If no subsumer was found, add the classifier to the population
        self.addClassifierToPopulation(cl)
     
        
    def increaseNumerositySum(self, nr):
        """ Increases recursively all numerositySum values in the set and all parent sets.
        This function should be called when the numerosity of a classifier in some set is increased in
        order to keep the numerosity sums of all sets and essentially the population up to date. """
        self.numerositySum = self.numerositySum + nr
        if self.parentSet != None:
            self.parentSet.increaseNumerositySum(nr)
            
            
    def addClassifierToPopulation(self,cl):
        """ Adds the classifier to the population and checks if an identical classifier exists.
        If an identical classifier exists, its numerosity is increased. """
        # set pop to the actual population
        pop=self
        while pop.parentSet!=None:
            pop = pop.parentSet
    
        oldcl=pop.getIdenticalClassifier(cl)
        
        if oldcl!=None:
            oldcl.addNumerosity(1)
            self.increaseNumerositySum(1)
        else:
            pop.addClassifier(cl)
          
          
    def getIdenticalClassifier(self, newCl):
        """ Looks for an identical classifier in the population.
        @param newCl The new classifier.
        @return Returns the identical classifier if found, null otherwise. """
        for cl in self.clSet:
            if newCl.equals(cl):
                return cl
        return None  
    
    #---------------------------------------------------------------------------------------
    def getAccuracySum(self):
        """  Returns the sum of the accuracy of all classifiers in the set.  """
        sum=0.0
        for cl in self.clSet:
            sum = sum + cl.getAccuracy() * cl.getNumerosity()
        return sum
    
    #---------------------------------------------------------------------------------------

    
    #NOTE don't make .gephi file here, do this externally
    #here just include co-occurence output file (stick with my original idea)
    def attCo_Occurence(self, outFile, env, exploreIter):
        """ Calculates pairwise attribute co-occurence througout all rules in the population - Based in part on Ambrose's code."""
        print "Calculating Co-occurence"
        
        comboList = []
        castList = [None,None,0] #att1, att2, co-occurence count
        count = 0
        dim = env.numAttributes
        for i in range(0,dim-1):
            for j in range(i+1,dim):
                comboList.append(copy.deepcopy(castList))
                comboList[count][0] = env.headerList[i]
                comboList[count][1] = env.headerList[j]   
                count += 1

        for cl in self.clSet:
            count = 0
            for i in range(dim-1):
                for j in range(i+1,dim):
                    if cl.condition[i] != cons.wild and cl.condition[j] != cons.wild:
                        if cons.attDisc == 'Spec':
                            comboList[count][2] += cl.numerosity
                        elif cons.attDisc == 'FWSpec':
                            comboList[count][2] += cl.numerosity*cl.fitness
                        elif cons.attDisc == 'AWSpec':
                            comboList[count][2] += cl.numerosity*cl.accuracy
                        else:
                            print "Co-occurence Calc. -- Specified attribute discovery not found."
                    count += 1
        
        
        #make tuples
        tupleList = []
        for i in comboList:
            tupleList.append((i[0],i[1],i[2]))

                
        sortedComboList = sorted(tupleList,key=lambda test: test[2], reverse=True)
        
        f = open(outFile + '_'+ str(exploreIter + 1)+ '_CO.txt', 'w')
        for i in range(len(sortedComboList)):
            for j in range(len(sortedComboList[0])): #att1, att2, count
                if j < len(sortedComboList[0])-1:
                    f.write(str(sortedComboList[i][j])+'\t')
                else:
                    f.write(str(sortedComboList[i][j])+'\n')
        f.close()
        
        """
        dim = env.numAttributes
        adj_m = np.zeros((dim, dim))
        for cl in self.clSet:
            inds = co_helper(cl.condition)
            for i in inds:
                for j in inds:
                    if cons.attDisc == 'Spec':
                        adj_m[i][j] += float(cl.numerosity)
                    elif cons.attDisc == 'FWSpec':
                        adj_m[i][j] += float(cl.numerosity*cl.fitness)
                    elif cons.attDisc == 'AWSpec':
                        adj_m[i][j] += float(cl.numerosity*cl.accuracy)
                    else:
                        print "Co-occurence Calc. -- Specified attribute discovery not found."
                    
        #Save an Adjacency Matrix
        f = open(outFile + '_'+ str(exploreIter + 1)+ '_COAdj.csv', 'w')
        for n in range(len(env.headerList)-1):
            f.write(';'+env.headerList[n])
        f.write('\n')
        for i in range(len(adj_mat)):
            f.write(env.headerList[i])
            for j in range(len(adj_mat[i])):
                f.write(';')
                f.write(str(adj_mat[i][j]))
            f.write('\n')
        f.close()
        

        
        adj_m = np.zeros((dim, dim))
        for datapoint in comboList:
            inds =  co_helper(datapoint)
        """  
##################################################################################################



    #Evaluation Methods---------------------------------------------------------------------------------------
    def runGenSum(self, env):
        """ Determines current Generality and Weighted Generality of population """  #Dictonary Updates
        self.attributeSpecList = []
        self.attributeAccList = []
        condition = []
        for i in range(env.getNumAttributes()):
            self.attributeSpecList.append(int(0))
            self.attributeAccList.append(int(0))
            
        genSum = 0.0
        fitGenSum = 0.0
        fitSum = 0.0 
        for cl in self.clSet:
            condition = cl.getCondition()
            
            #ALT RULE REP
            indCondGen = 0.0
            for j in range(len(condition)):
                if condition[j] != cons.wild:
                    self.attributeSpecList[j] += cl.getNumerosity() #Attribute Generality takes numerosity into acount!!!
                    self.attributeAccList[j] += (cl.getNumerosity()*cl.getAccuracy())
                else:
                    indCondGen += 1
            indCondGen = indCondGen/float(len(condition))
            
            cl.setClassifierGenerality(indCondGen) 
            genSum = genSum + indCondGen * cl.getNumerosity()
            fitGenSum = fitGenSum + indCondGen * cl.getNumerosity() * cl.getFitness()
            fitSum = fitSum + cl.getNumerosity() * cl.getFitness()

        if self.numerositySum == 0:
            self.generality = 'NA'
        else:
            self.generality = genSum / float(self.numerositySum)
        if fitSum == 0:
            self.weightedGenerality = 'NA'
        else:
            self.weightedGenerality = fitGenSum / float(fitSum)  #zero division error (no fitness!!!


    ####ALTERNATE NOISE HANDELING############################################################################################
    def getSpecSum(self,env):
        """ For RAIN: independently calculates the attributes generality sums """
        self.attributeSpecList = []
        for i in range(env.getNumAttributes()):
            self.attributeSpecList.append(int(0))
            
        for cl in self.clSet:
            condition = cl.getCondition()
            for j in range(len(condition)):
                if condition[j] != cons.wild:
                    self.attributeSpecList[j] += (1*cl.getNumerosity()) #Attribute Generality takes numerosity into acount!!!

        return self.attributeSpecList
      
      
    def getFWSpecSum(self,env):
        """ For RAIN: independently calculates the attributes fitness weighted generality sums """
        self.attributeFWSpecList = []
        for i in range(env.getNumAttributes()):
            self.attributeFWSpecList.append(int(0))
            
        for cl in self.clSet:
            condition = cl.getCondition()
            for j in range(len(condition)):
                if condition[j] != cons.wild:
                    self.attributeFWSpecList[j] += (cl.getNumerosity()*cl.getFitness())

        return self.attributeFWSpecList
              
              
    def getAWSpecSum(self,env):
        """ For RAIN: independently calculates the attributes specified accuracy sums -i.e. when an attribute is specified in a rule we add that rule's
        accuracy to the attributes collective sum."""
        self.attributeAccList = []
        for i in range(env.getNumAttributes()):
            self.attributeAccList.append(int(0))
            
        for cl in self.clSet:
            condition = cl.getCondition()
            for j in range(len(condition)):
                if condition[j] != cons.wild:
                    self.attributeAccList[j] += (cl.getNumerosity()*cl.getAccuracy()) 

        return self.attributeAccList
    ######################################################################################################################################
    
    
    def getPopTrack(self, acc, exploreIter, timer):
        """ Returns a formated output string to be printed to the Learn Track output file. """
        #HEADER FOR POP_TRACK: "Explore_Iteration\tMacroPopSize\tMicroPopSize\tLocalAccuracy\tGenerality\tWeightedGenerality\tTime(min)\n"
        trackString = str(exploreIter)+ "\t" + str(len(self.clSet)) + "\t" + str(self.numerositySum) + "\t" + str(acc) + "\t" + str(self.generality) + "\t" + str(self.weightedGenerality) + "\t" + str(timer.endGlobalTimer())+ "\n"
        print ("Iter: " + str(exploreIter) + "\t MacroPop: " + str(len(self.clSet))+ "\t MicroPop: " + str(self.numerositySum)+ "\t Acc: " + str(acc) + "\t Gen: " + str(self.generality)+ "\t WGen: " + str(self.weightedGenerality)+ "\t Time: " + str(timer.endGlobalTimer()))
        return trackString
        
        
    def writePopTrack(self, outFile, trainEval, testEval, env, timer, exploreIter):
        """ Makes output text file which includes all of the parameter settings used in the run as well as all of the evaluation stats including Time Track Output. """
        try:  
            popStatsOut = open(outFile + '_'+ str(exploreIter + 1)+'_PopStats.txt','w') # Outputs Population run stats
            print "Writing Population Statistical Summary File"
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        
        #Evaluation of pop
        popStatsOut.write("Performance Stats:\n")
        popStatsOut.write("Training_Accuracy\tTesting_Accuracy\tNoMatchFreq_Train\tNoMatchFreq_Test\n")
        popStatsOut.write(str(trainEval[0])+"\t")
        if cons.crossVal:
            popStatsOut.write(str(testEval[0])+"\t")
            popStatsOut.write(str(trainEval[1]) +"\t")
            popStatsOut.write(str(testEval[1])+"\n\n") 
        else:
            popStatsOut.write("NA\t")
            popStatsOut.write(str(trainEval[1]) +"\t")
            popStatsOut.write("NA\n\n")
        
        popStatsOut.write("Population Characterization:\n")  
        if self.numerositySum == 0:
            aveacc = 0
            avefit = 0
        else:
            aveacc = self.getAccuracySum()/self.numerositySum
            avefit = self.getFitnessSum()/self.numerositySum
        popStatsOut.write("MacroPopSize\tMicroPopSize\tGenerality\tWeightedGenerality\tAverageAccuracy\tAverageFitness\n") 
        popStatsOut.write(str(len(self.clSet))+"\t"+ str(self.numerositySum)+"\t"+str(self.generality)+"\t"+str(self.weightedGenerality)+"\t"+str(aveacc)+"\t"+str(avefit)+"\n\n")

        popStatsOut.write("SpecificitySum:\n")
        headList = env.getHeaderList()
        for i in range(len(headList)-1):    # Added the -1 to get rid of the Class Header
            if i < len(headList)-2:
                popStatsOut.write(str(headList[i])+"\t")
            else:
                popStatsOut.write(str(headList[i])+"\n")
                
        # Prints out the Spec Count for each attribute 
        for i in range(len(self.attributeSpecList)):
            if i < len(self.attributeSpecList)-1:
                popStatsOut.write(str(self.attributeSpecList[i])+"\t")
            else:
                popStatsOut.write(str(self.attributeSpecList[i])+"\n")  
                    
        popStatsOut.write("\nAccuracySum:\n")
        headList = env.getHeaderList()
        for i in range(len(headList)-1):    # Added the -1 to get rid of the Class Header
            if i < len(headList)-2:
                popStatsOut.write(str(headList[i])+"\t")
            else:
                popStatsOut.write(str(headList[i])+"\n")
                
        # Prints out the Accuracy Based Spec Count for each attribute 
        for i in range(len(self.attributeAccList)):
            if i < len(self.attributeAccList)-1:
                popStatsOut.write(str(self.attributeAccList[i])+"\t")
            else:
                popStatsOut.write(str(self.attributeAccList[i])+"\n")    
                    
        #Time Track      
        popStatsOut.write("\nTime Track(in minutes):\n") 
        popStatsOut.write(timer.reportTimes())

        #Run Parameters and constants
        popStatsOut.write("\nRun Parameters::\n\n") 
        popStatsOut.write(cons.getParamSummary())

        try:
            popStatsOut.close()
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise   
                
                
    def writePop(self, outFile, env, exploreIter):
        """ Writes a tab delimited text file specifying the evolved rule population, including conditions, actions, and all rule parameters. """
        try:  
            rulePopOut = open(outFile + '_'+ str(exploreIter + 1)+'_RulePop.txt','w') # Outputs tab delimited text file of rule population and respective rule stats
            print "Writing Population as Data File"
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise
        
        headList = env.getHeaderList()
        for i in range(len(headList)-1):    # Added the -1 to get rid of the Class Header
            rulePopOut.write(str(headList[i])+"\t")

        rulePopOut.write("Action\tFitness\tAccuracy\tCorrectTrack\tExperience\tNumerosity\tActionSetSize\tTimeStamp\tGenerality\tTrainMatch\tTrainCorrect\tTrainAccuracy\tTestMatch\tTestCorrect\tTestAccuracy\tInitTimeStamp\n")
        #Write each classifier
        for cl in self.clSet:
            rulePopOut.write(str(cl.printClassifier()))  #implement printClassifier to return string with tabs formated for outputfile with \n.

        try:
            rulePopOut.close() 
        except IOError, (errno, strerror):
            print ("I/O error(%s): %s" % (errno, strerror))
            raise   
             
    def recalculateNumerositySum(self):
        self.numerositySum = 0
        for cl in self.clSet:
            self.numerositySum = self.numerositySum + cl.getNumerosity()
        
        
    #Uncertain if Needed--------------------------------------------------------------------------------------
    def confirmClassifiersInSet(self):
        """ Updates the numerositySum of the set and deletes all classifiers with numerosity 0. """
        borrados = 0
        for cl in self.clSet[:]:
            if cl.getNumerosity()==0:
                self.clSet.remove(cl)
                borrados = borrados  + 1
            else:
                self.numerositySum = self.numerositySum + cl.getNumerosity()

    
    def getNumerositySum(self):
        """ Returns the number of micro-classifiers in the set. """
        return self.numerositySum

        
    def getSize(self):
        """ Returns the number of macro-classifiers in the set. """
        return len(self.clSet)  # true size
    #---------------------------------------------------------------------------------------
    
    def checkNumerositySum(self):
        """ DEBUG METHOD: recalculates numerositySum of the set, directly from the classifiers """
        numSum = 0
        for i in range(len(self.clSet)):
            numSum += self.clSet[i].numerosity
        return numSum
